/* @exclude */
/* jshint -W069 */
/* jshint -W041 */
/* jshint -W018 */
/* global composite */
/* @endexclude */

/**
 * Sorting, filtering and pagination data and methods associated with the available component options.
 */
this.Component_Options_Model = function( component, opts ) {

	var self  = component;
	var Model = Backbone.Model.extend( {

		available_options_data: [],
		xhr: false,

		initialize: function() {

			this.available_options_data = self.find_options_data();

			var available_options = [];

			if ( this.available_options_data.length > 0 ) {
				available_options = _.pluck( this.available_options_data, 'option_id' );
			}

			var params = {
				filters: self.find_active_filters(),
				orderby: self.find_order_by(),
				page:    self.find_pagination_param( 'page' ),
				pages:   self.find_pagination_param( 'pages' ),

				/*
				 * Available component options, including the current selection, but excluding the empty '' option.
				 */
				available_options: available_options,

				/*
				 * Valid products and variations, including the current selection and the empty '' option.
				 */
				active_options: available_options.slice(),

				/*
				 * Products and variations state in current view, including the empty '' option. The current selection is excluded if not in view.
				 */
				options_state: { active: _.pluck( _.where( this.available_options_data, { is_in_view: true } ), 'option_id' ), inactive: [] },

				/**
				 * 'compat_group' scenarios when the options state was calculated.
				 */
				options_in_scenarios: composite.scenarios.clean_masked_scenarios( composite.scenarios.get_scenarios_by_type( 'compat_group' ), self.component_id )
			};

			this.set( params );

			/**
			 * Refresh options state when a selection changes.
			 */
			composite.actions.add_action( 'component_selection_changed', this.component_selection_changed_handler, 15, this );

			/**
			 * Ensure options state is refreshed when the available options of this component change.
			 */
			composite.actions.add_action( 'available_options_changed_' + self.step_id, this.available_options_changed_handler, 10, this );

			if ( composite.settings.layout !== 'single' ) {

				/*
				 * Reset invalid product/variation selections when transitioning to this step.
				 */
				composite.actions.add_action( 'active_step_changed_' + self.step_id, this.active_step_changed_handler, 10, this );
			}

		},

		active_step_changed_handler: function() {

			if ( ! composite.is_initialized ) {
				return;
			}

			this.refresh_options_state();
		},

		available_options_changed_handler: function() {

			this.refresh_options_state();
		},

		component_selection_changed_handler: function( step ) {

			// Only update options if change happened in preceding step.
			if ( composite.settings.layout && self.step_index < step.step_index ) {
				return false;
			}

			this.refresh_options_state();
		},

		reload_options_on_scenarios_change: function() {

			var reload = false;

			if ( self.hide_disabled_products() && self.has_options_style( 'thumbnails' ) ) {
				if ( self.get_max_results() > self.get_results_per_page() ) {
					if ( false === self.append_results() ) {
						reload = true;
					} else if ( _.pluck( _.where( this.available_options_data, { is_in_view: true } ), 'option_id' ).length < self.get_max_results() ) {
						reload = true;
					}
				}
			}

			return reload;
		},

		request_options: function( params, request_type ) {

			var model = this;

			// Page will be updated after data has been fetched.
			this.set( _.omit( params, 'page' ) );

			var data = {
				action:               'woocommerce_show_component_options',
				component_id:         self.component_id,
				composite_id:         composite.composite_id,
				load_page:            params.page ? params.page : 1,
				selected_option:      self.get_selected_product( false ),
				filters:              this.get( 'filters' ),
				orderby:              this.get( 'orderby' ),
				options_in_scenarios: this.reload_options_on_scenarios_change() ? this.get( 'options_in_scenarios' ) : [],
			};

			if ( this.xhr ) {
				this.xhr.abort();
			}

			// Get component options via ajax.
			this.xhr = $.post( composite.get_ajax_url( data.action ), data, function( response ) {

				// Trigger 'component_options_data_loaded' event.
				model.trigger( 'component_options_data_loaded', response, request_type );

				if ( 'success' === response.result ) {

					if ( 'reload' === request_type ) {

						// Update component options data.
						model.available_options_data = response.options_data;

						// Update component scenario data.
						composite.scenarios.set_component_scenario_data( self.component_id, response.scenario_data );

						// Update component pagination data.
						model.set( response.pagination_data );

						// Update available options.
						model.refresh_options( _.pluck( model.available_options_data, 'option_id' ) );

					} else if ( 'append' === request_type ) {

						// Merge existing with new component options data, after adding an 'is_appended' prop to the new data.
						model.available_options_data = _.union( _.where( model.available_options_data, { is_in_view: true } ), _.map( response.options_data, function( option_data ) { return _.extend( option_data, { is_appended: true } ); } ) );

						// Merge component scenario data.
						composite.scenarios.merge_component_scenario_data( self.component_id, response.scenario_data );

						// Update component pagination data.
						model.set( response.pagination_data );

						// Update available options.
						model.refresh_options( _.pluck( model.available_options_data, 'option_id' ) );

						// Remove 'is_appended' prop from appended data.
						model.available_options_data = _.map( model.available_options_data, function( option_data ) { return _.omit( option_data, 'is_appended' ); } );
					}

				} else {
					window.alert( response.message );
				}

				// Run 'component_options_loaded' action - @see WC_CP_Actions_Dispatcher class reference.
				composite.actions.do_action( 'component_options_loaded', [ self ] );

			}, 'json' );
		},

		get_option_data: function( option_id, key ) {

			var data = null;

			if ( option_id !== '' && this.available_options_data.length > 0 ) {
				$.each( this.available_options_data, function( index, option_data ) {
					if ( parseInt( option_data.option_id, 10 ) === parseInt( option_id, 10 ) ) {
						if ( key ) {
							data = typeof( option_data[ key ] ) !== 'undefined' ? option_data[ key ] : null;
						} else {
							data = option_data;
						}
						return false;
					}
				} );
			}

			return data;
		},

		set_option_data: function( option_id, data, key ) {

			var model = this;

			if ( option_id !== '' && this.available_options_data.length > 0 ) {
				$.each( this.available_options_data, function( index, option_data ) {
					if ( parseInt( option_id, 10 ) === parseInt( option_data.option_id, 10 ) ) {
						if ( key ) {
							model.available_options_data[ index ][ key ] = data;
						} else {
							model.available_options_data[ index ] = data;
						}
						return false;
					}
				} );
			}
		},

		refresh_options: function( options ) {

			composite.console_log( 'debug:models', '\nUpdating "' + self.get_title() + '" options: ' + _.map( options, function( num ) { return num === '' ? '0' : num; } ) );

			composite.debug_indent_incr();

			if ( _.isEqual( this.get( 'available_options' ), options ) ) {
				// Refresh options state if options have been refreshed but the new set is equal to the old: Edge case fix for when the 'is_in_view' property of an existing option changes in the new set.
				this.refresh_options_state();
			} else {
				this.set( { available_options: options } );
			}
			composite.debug_indent_decr();
		},

		refresh_options_state: function() {

			/*
			 * 1. Update active options.
			 */

			composite.console_log( 'debug:models', '\nUpdating "' + self.get_title() + '" options state...' );

			composite.debug_indent_incr();

			var active_options          = [],
				active_scenarios        = [],
				options_state           = { active: [], inactive: [] },
				component_id            = self.component_id,
				scenario_data           = composite.scenarios.get_scenario_data().scenario_data,
				item_scenario_data      = scenario_data[ component_id ],
				is_optional             = false,
				invalid_product_found   = false,
				invalid_variation_found = false;

			// Get active 'compat_group' scenarios up to this component.
			active_scenarios = composite.scenarios.filter_scenarios_by_type( composite.scenarios.calculate_active_scenarios( self, true, true ), 'compat_group' );

			composite.console_log( 'debug:models', '\nReference scenarios: [' + active_scenarios + ']' );
			composite.console_log( 'debug:models', 'Removing any scenarios where the current component is masked...' );

			active_scenarios = composite.scenarios.clean_masked_scenarios( active_scenarios, component_id );

			// Enable all if all active scenarios ignore this component.
			if ( active_scenarios.length === 0 ) {
				active_scenarios.push( '0' );
			}

			composite.console_log( 'debug:models', '\nUpdating \'Component_Options_Model\': "' + self.get_title() + '", Attribute: "options_in_scenarios"...' );

			if ( ! _.isEqual( this.get( 'options_in_scenarios' ), active_scenarios ) ) {
				composite.console_log( 'debug:models', '\nActive options scenarios changed - [' + this.get( 'options_in_scenarios' ) + '] => [' + active_scenarios + '].\n' );
			} else {
				composite.console_log( 'debug:models', '\nActive options scenarios unchanged - [' + this.get( 'options_in_scenarios' ) + '] => [' + active_scenarios + '].\n' );
			}

			composite.debug_indent_incr();
			this.set( { options_in_scenarios: active_scenarios } );
			composite.debug_indent_decr();

			/*
			 * Set component 'optional' status by adding the '' product ID to the 'active_options' array.
			 */

			if ( 0 in item_scenario_data ) {

				var optional_in_scenarios = item_scenario_data[ 0 ];

				for ( var s = 0; s < optional_in_scenarios.length; s++ ) {

					var optional_in_scenario_id = optional_in_scenarios[ s ];

					if ( $.inArray( optional_in_scenario_id, active_scenarios ) > -1 ) {
						is_optional = true;
						break;
					}
				}

				options_state.inactive.push( '' );

			} else if ( false === self.is_visible() ) {
				is_optional = true;
			}

			if ( is_optional ) {
				composite.console_log( 'debug:models', 'Component set as optional.' );
				active_options.push( '' );
				options_state.active.push( '' );
			}

			/*
			 * Add compatible products to the 'active_options' array.
			 */
			$.each( this.available_options_data, function( index, option_data ) {

				var product_id           = option_data.option_id,
					product_in_scenarios = ( product_id in item_scenario_data ) ? item_scenario_data[ product_id ] : [],
					is_compatible        = false;

				composite.console_log( 'debug:models', 'Updating selection #' + product_id + ':' );
				composite.console_log( 'debug:models', '	Selection in scenarios: [' + product_in_scenarios + ']' );

				for ( var i = 0; i < product_in_scenarios.length; i++ ) {

					var scenario_id = product_in_scenarios[ i ];

					if ( $.inArray( scenario_id, active_scenarios ) > -1 ) {
						is_compatible = true;
						break;
					}
				}

				if ( is_compatible ) {
					composite.console_log( 'debug:models', '	Selection enabled.' );
					active_options.push( product_id );

					if ( option_data.is_in_view ) {
						options_state.active.push( product_id );
					}

				} else {

					composite.console_log( 'debug:models', '	Selection disabled.' );

					if ( option_data.is_in_view ) {
						options_state.inactive.push( product_id );
					}

					if ( self.get_selected_product( false ) === product_id ) {

						invalid_product_found = true;

						if ( invalid_product_found ) {
							composite.console_log( 'debug:models', '	--- Selection invalid.' );
						}
					}
				}
			} );

			/*
			 * Disable incompatible variations.
			 */

			if ( self.get_selected_product_type() === 'variable' ) {

				var chosen_variation = self.get_selected_variation(),
					variation_data   = self.component_selection_model.get_available_variations_data(),
					variation_id,
					variation_in_scenarios,
					is_compatible;

				composite.console_log( 'debug:models', '	Checking variations...' );

				if ( chosen_variation > 0 ) {
					composite.console_log( 'debug:models', '		--- Stored variation is #' + chosen_variation );
				}

				/*
				 * Update model.
				 */
				for ( var i = 0; i < variation_data.length; i++ ) {

					variation_id  = variation_data[ i ].variation_id.toString();
					is_compatible = false;

					variation_in_scenarios = ( variation_id in item_scenario_data ) ? item_scenario_data[ variation_id ] : [];

					composite.console_log( 'debug:models', '		Checking variation #' + variation_id + ':' );
					composite.console_log( 'debug:models', '		Selection in scenarios: [' + variation_in_scenarios + ']' );

					for ( var k = 0; k < variation_in_scenarios.length; k++ ) {

						var scenario_id = variation_in_scenarios[ k ];

						if ( $.inArray( scenario_id, active_scenarios ) > -1 ) {
							is_compatible = true;
							break;
						}
					}

					if ( is_compatible ) {

						composite.console_log( 'debug:models', '		Variation enabled.' );

						active_options.push( variation_id );
						options_state.active.push( variation_id );

					} else {

						composite.console_log( 'debug:models', '		Variation disabled.' );

						options_state.inactive.push( variation_id.toString() );

						if ( self.get_selected_variation( false ).toString() === variation_id ) {

							invalid_variation_found = true;

							if ( invalid_variation_found ) {
								composite.console_log( 'debug:models', '		--- Selection invalid.' );
							}
						}
					}
				}
			}

			composite.console_log( 'debug:models', 'Done.\n' );

			composite.debug_indent_decr();

			var maybe_update_model = true;

			/*
			 * 2. Check selections.
			 */

			if ( composite.settings.layout === 'single' || self.is_current() ) {

				composite.console_log( 'debug:models', '\nChecking current "' + self.get_title() + '" selections:' );

				if ( invalid_variation_found ) {

					maybe_update_model = false;

					composite.console_log( 'debug:models', '\nVariation selection invalid - resetting...\n\n' );

					composite.debug_indent_incr();

					self.component_selection_view.resetting_variation = true;
					self.$component_summary_content.find( '.reset_variations' ).trigger( 'click' );
					self.component_selection_view.resetting_variation = false;

					composite.debug_indent_decr();

				} else if ( invalid_product_found ) {
					composite.console_log( 'debug:models', '\nProduct selection invalid.\n\n' );
				} else  {
					composite.console_log( 'debug:models', '...looking good!' );
				}
			}

			/*
			 * 3. Update model.
			 */

			if ( maybe_update_model ) {

				if ( ! _.isEqual( this.get( 'active_options' ), active_options ) ) {
					composite.console_log( 'debug:models', '\nActive options changed - [' + _.map( this.get( 'active_options' ), function( num ) { return num === '' ? '0' : num; } ) + '] => [' + _.map( active_options, function( num ) { return num === '' ? '0' : num; } ) + '].\n' );
				} else {
					composite.console_log( 'debug:models', '\nActive options unchanged - [' + _.map( this.get( 'active_options' ), function( num ) { return num === '' ? '0' : num; } ) + '] => [' + _.map( active_options, function( num ) { return num === '' ? '0' : num; } ) + '].\n' );
				}

				// Set active options.
				composite.debug_indent_incr();
				this.set( { active_options: active_options } );
				composite.debug_indent_decr();


				if ( ! _.isEqual( this.get( 'options_state' ), options_state ) ) {
					composite.console_log( 'debug:models', '\nOptions state changed.\n' );
				} else {
					composite.console_log( 'debug:models', '\nOptions state unchanged.\n' );
				}

				// Set active options in view.
				composite.debug_indent_incr();
				this.set( { options_state: options_state } );
				composite.debug_indent_decr();
			}
		}

	} );

	var obj = new Model( opts );
	return obj;
};
