/* @exclude */
/* jshint -W069 */
/* jshint -W041 */
/* jshint -W018 */
/* global composite */
/* @endexclude */

/**
 * Composite product data model for storing validation, pricing, availability and quantity data.
 */
this.Composite_Data_Model = function( opts ) {

	var Model = Backbone.Model.extend( {

		price_data:  composite.$composite_data.data( 'price_data' ),
		$nyp:        false,

		initialize: function() {

			var params = {
				passes_validation:     true,
				validation_messages:   [],
				status_messages:       [],
				is_in_stock:           true,
				stock_statuses:        [],
				totals:                { price: '', regular_price: '', price_incl_tax: '', price_excl_tax: '' }
			};

			$.each( composite.get_components(), function( index, component ) {
				params[ 'component_' + component.component_id + '_totals' ] = { price: '', regular_price: '', price_incl_tax: '', price_excl_tax: '' };
			} );

			this.set( params );

			// Price suffix data.
			this.price_data.suffix_exists              = wc_composite_params.price_display_suffix !== '';
			this.price_data.suffix_contains_price_incl = wc_composite_params.price_display_suffix.indexOf( '{price_including_tax}' ) > -1;
			this.price_data.suffix_contains_price_excl = wc_composite_params.price_display_suffix.indexOf( '{price_excluding_tax}' ) > -1;

			/**
			 * Update model totals when the nyp price changes.
			 */
			composite.actions.add_action( 'component_nyp_changed', this.nyp_changed_handler, 10, this );

			/**
			 * Update model totals state when a new component quantity is selected.
			 */
			composite.actions.add_action( 'component_quantity_changed', this.quantity_changed_handler, 20, this );

			/**
			 * Update model totals state when a new selection is made.
			 */
			composite.actions.add_action( 'component_selection_changed', this.selection_changed_handler, 30, this );

			/**
			 * Update totals when the contents of an existing selection change.
			 */
			composite.actions.add_action( 'component_selection_content_changed', this.selection_content_changed_handler, 30, this );

			/**
			 * Update model availability state in response to component changes.
			 */
			composite.actions.add_action( 'component_availability_changed', this.availability_changed_handler, 10, this );

			/**
			 * Update model validation state when a step validation model state changes.
			 */
			composite.actions.add_action( 'component_validation_message_changed', this.validation_status_changed_handler, 10, this );
			composite.actions.add_action( 'component_validation_status_changed', this.validation_status_changed_handler, 10, this );

			/**
			 * Update a single summary view element price when its totals change.
			 */
			composite.actions.add_action( 'component_totals_changed', this.component_totals_changed_handler, 10, this );

			/**
			 * Update composite totals when a new NYP price is entered at composite level.
			 */
			var $nyp = composite.$composite_data.find( '.nyp' );

			if ( $nyp.length > 0 ) {

				this.$nyp                       = $nyp;
				this.price_data[ 'base_price' ] = $nyp.data( 'price' );

				composite.$composite_data.on( 'woocommerce-nyp-updated-item', { model: this }, function( event ) {

					var model = event.data.model;

					model.price_data[ 'base_price' ]         = model.$nyp.data( 'price' );
					model.price_data[ 'base_regular_price' ] = model.$nyp.data( 'price' );

					model.calculate_totals();
				} );
			}
		},

		/**
		 * Initializes the model and prepares data for consumption by views.
		 */
		init: function() {

			composite.console_log( 'debug:models', '\nInitializing composite data model...' );
			composite.debug_indent_incr();

			this.update_validation();
			this.update_totals();
			this.update_availability();

			composite.debug_indent_decr();
		},

		/**
		 * Updates component totals when a nyp price change event is triggered.
		 */
		nyp_changed_handler: function( component ) {

			if ( ! composite.is_initialized ) {
				return false;
			}

			this.update_totals( component );
		},

		/**
		 * Updates model totals state.
		 */
		selection_changed_handler: function() {

			if ( ! composite.is_initialized ) {
				return false;
			}

			this.update_validation();
			this.update_totals();
		},

		/**
		 * Updates model availability state.
		 */
		availability_changed_handler: function() {

			if ( ! composite.is_initialized ) {
				return false;
			}

			this.update_availability();
		},

		/**
		 * Updates model totals state.
		 */
		selection_content_changed_handler: function() {

			if ( ! composite.is_initialized ) {
				return false;
			}

			this.update_validation();
			this.update_totals();
		},

		/**
		 * Updates model totals state.
		 */
		quantity_changed_handler: function( component ) {

			if ( ! composite.is_initialized ) {
				return false;
			}

			this.update_totals( component );
		},

		/**
		 * Updates model validation state when the state of a step validation model changes.
		 */
		validation_status_changed_handler: function() {
			this.update_validation();
		},

		// Updates totals when component subtotals change.
		component_totals_changed_handler: function() {
			this.calculate_totals();
		},

		/**
		 * Updates the validation state of the model.
		 */
		update_validation: function() {

			var messages = [],
				status   = this.get_validation_status();

			if ( this.is_purchasable() ) {
				messages = this.get_validation_messages();
			} else {
				messages.push( wc_composite_params.i18n_unavailable_text );
			}

			composite.console_log( 'debug:models', '\nUpdating \'Composite_Data_Model\' validation state... Attribute count: "validation_messages": ' + messages.length + ', Attribute: "passes_validation": ' + ( messages.length === 0 ).toString() );

			composite.debug_indent_incr();
			this.set( { validation_messages: messages, passes_validation: status } );
			composite.debug_indent_decr();
		},

		/**
		 * Get aggregate validation status.
		 */
		get_validation_status: function() {

			var passes_validation = true;

			if ( ! this.is_purchasable() ) {
				return false;
			}

			$.each( composite.get_steps(), function( step_index, step ) {
				if ( ! step.step_validation_model.get( 'passes_validation' ) ) {
					passes_validation = false;
					return false;
				}
			} );

			return passes_validation;
		},

		/**
		 * Get all validation messages grouped by source. Messages added from the 'Review' step are displayed individually.
		 */
		get_validation_messages: function() {

			var validation_messages = [];

			$.each( composite.get_steps(), function( step_index, step ) {

				var source = step.get_title();

				$.each( step.get_validation_messages( 'composite' ), function( message_index, message ) {

					if ( step.is_review() ) {
						validation_messages.push( { sources: false, content: message.toString() } );
					} else {
						var appended = false;

						if ( validation_messages.length > 0 ) {
							$.each( validation_messages, function( id, msg ) {
								if ( msg.content === message ) {
									var sources_new = msg.sources;
									var content_new = msg.content;
									sources_new.push( source );
									validation_messages[ id ] = { sources: sources_new, content: content_new };
									appended = true;
									return false;
								}
							} );
						}

						if ( ! appended ) {
							validation_messages.push( { sources: [ source ], content: message.toString() } );
						}
					}

				} );

			} );

			var messages = [];

			if ( validation_messages.length > 0 ) {
				$.each( validation_messages, function( id, msg ) {
					if ( msg.sources === false ) {
						messages.push( msg.content );
					} else {
						var sources = wc_cp_join( msg.sources );
						messages.push( wc_composite_params.i18n_validation_issues_for.replace( '%c', sources ).replace( '%e', msg.content ) );
					}
				} );
			}

			// Pass through 'composite_validation_messages' filter - @see WC_CP_Filters_Manager class.
			messages = composite.filters.apply_filters( 'composite_validation_messages', [ messages ] );

			return messages;
		},

		/**
		 * True if the product is purchasable.
		 */
		is_purchasable: function() {

			if ( this.price_data[ 'is_purchasable' ] === 'no' ) {
				return false;
			}

			return true;
		},

		/**
		 * Get the composite quantity.
		 */
		get_quantity: function() {
			return composite.$composite_quantity.length > 0 ? parseInt( composite.$composite_quantity.val(), 10 ) : 1;
		},

		/**
		 * Updates model availability state.
		 */
		update_availability: function() {

			var stock_statuses = [],
				is_in_stock    = true;

			$.each( composite.get_components(), function( index, component ) {
				stock_statuses.push( component.step_validation_model.get( 'is_in_stock' ) );
			} );

			is_in_stock = _.contains( stock_statuses, false ) ? false : true;

			composite.console_log( 'debug:models', '\nUpdating \'Composite_Data_Model\' availability... Attribute: "stock_statuses": ' + stock_statuses.toString() + ', Attribute: "is_in_stock": ' + is_in_stock.toString() );

			composite.debug_indent_incr();
			this.set( {
				stock_statuses: stock_statuses,
				is_in_stock:    is_in_stock,
			} );
			composite.debug_indent_decr();
		},

		/**
		 * Calculates and updates model subtotals.
		 */
		update_totals: function( component ) {

			var model = this;

			composite.console_log( 'debug:models', '\nUpdating \'Composite_Data_Model\' totals...' );

			composite.debug_indent_incr();

			if ( typeof( component ) === 'undefined' ) {

				$.each( composite.get_components(), function( index, component ) {
					model.update_component_prices( component );
				} );

				this.calculate_subtotals();

			} else {
				this.update_component_prices( component );
				this.calculate_subtotals( component );
			}

			composite.debug_indent_decr();
		},

		/**
		 * Calculates totals by applying tax ratios to raw prices.
		 */
		get_taxed_totals: function( price, regular_price, tax_ratios, qty ) {

			qty = typeof( qty ) === 'undefined' ? 1 : qty;

			var tax_ratio_incl = tax_ratios && typeof( tax_ratios.incl ) !== 'undefined' ? Number( tax_ratios.incl ) : false,
				tax_ratio_excl = tax_ratios && typeof( tax_ratios.excl ) !== 'undefined' ? Number( tax_ratios.excl ) : false,
				totals         = {
					price:          qty * price,
					regular_price:  qty * regular_price,
					price_incl_tax: qty * price,
					price_excl_tax: qty * price
				};

			if ( tax_ratio_incl && tax_ratio_excl ) {

				totals.price_incl_tax = wc_cp_number_round( totals.price * tax_ratio_incl );
				totals.price_excl_tax = wc_cp_number_round( totals.price * tax_ratio_excl );

				if ( wc_composite_params.tax_display_shop === 'incl' ) {
					totals.price         = totals.price_incl_tax;
					totals.regular_price = wc_cp_number_round( totals.regular_price * tax_ratio_incl );
				} else {
					totals.price         = totals.price_excl_tax;
					totals.regular_price = wc_cp_number_round( totals.regular_price * tax_ratio_excl );
				}
			}

			return totals;
		},

		/**
		 * Adds model subtotals and calculates model totals.
		 */
		calculate_totals: function( price_data_array ) {

			var model      = this,
				price_data = typeof( price_data_array ) === 'undefined' ? model.price_data : price_data_array;

			composite.console_log( 'debug:models', '\nAdding totals...' );

			var totals = {
				price:          price_data[ 'base_price_totals' ].price,
				regular_price:  price_data[ 'base_price_totals' ].regular_price,
				price_incl_tax: price_data[ 'base_price_totals' ].price_incl_tax,
				price_excl_tax: price_data[ 'base_price_totals' ].price_excl_tax
			};

			price_data[ 'base_display_price' ] = totals.price;

			$.each( composite.get_components(), function( index, component ) {

				var component_totals = typeof( price_data_array ) === 'undefined' ? model.get( 'component_' + component.component_id + '_totals' ) : price_data_array[ 'component_' + component.component_id + '_totals' ];

				totals.price          += component_totals.price;
				totals.regular_price  += component_totals.regular_price;
				totals.price_incl_tax += component_totals.price_incl_tax;
				totals.price_excl_tax += component_totals.price_excl_tax;
			} );

			// Pass through 'composite_totals' filter - @see WC_CP_Filters_Manager class.
			totals = composite.filters.apply_filters( 'composite_totals', [ totals ] );

			if ( typeof( price_data_array ) === 'undefined' ) {
				composite.debug_indent_incr();
				this.set( { totals: totals } );
				composite.debug_indent_decr();
			}

			return totals;
		},

		/**
		 * Calculates composite subtotals (component totals) and updates the component totals attributes on the model when the calculation is done on the client side.
		 * For components that require a server-side calculation of incl/excl tax totals, a request is prepared and submitted in order to get accurate values.
		 */
		calculate_subtotals: function( triggered_by, price_data_array, qty ) {

			var model      = this,
				price_data = typeof( price_data_array ) === 'undefined' ? model.price_data : price_data_array;

			qty          = typeof( qty ) === 'undefined' ? 1 : parseInt( qty, 10 );
			triggered_by = typeof( triggered_by ) === 'undefined' ? false : triggered_by;

			// Base.
			if ( false === triggered_by ) {

				var base_price            = Number( price_data[ 'base_price' ] ),
					base_regular_price    = Number( price_data[ 'base_regular_price' ] ),
					base_price_tax_ratios = price_data[ 'base_price_tax_ratios' ];

				price_data[ 'base_price_totals' ] = this.get_taxed_totals( base_price, base_regular_price, base_price_tax_ratios, qty );
			}

			// Components.
			$.each( composite.get_components(), function( index, component ) {

				if ( false !== triggered_by && triggered_by.component_id !== component.component_id ) {
					return true;
				}

				var component_qty = price_data[ 'quantities' ][ component.component_id ] * qty,
					totals        = model.calculate_component_subtotals( component, price_data, component_qty );

				if ( typeof( price_data_array ) === 'undefined' ) {

					composite.console_log( 'debug:models', 'Updating \'Composite_Data_Model\' component totals... Attribute: "component_' + component.component_id + '_totals".' );

					composite.debug_indent_incr();
					model.set( 'component_' + component.component_id + '_totals', totals );
					composite.debug_indent_decr();

				} else {
					price_data[ 'component_' + component.component_id + '_totals' ] = totals;
				}

			} );

			if ( typeof( price_data_array ) !== 'undefined' ) {
				return price_data;
			}
		},

		/**
		 * Calculates component subtotals on the client side.
		 */
		calculate_component_subtotals: function( component, price_data_array, qty ) {

			var model         = this,
				price_data    = typeof( price_data_array ) === 'undefined' ? model.price_data : price_data_array,
				product_id    = component.get_selected_product_type() === 'variable' ? component.get_selected_variation( false ) : component.get_selected_product( false ),
				tax_ratios    = price_data[ 'price_tax_ratios' ][ component.component_id ],
				regular_price = price_data[ 'regular_prices' ][ component.component_id ] + price_data[ 'addons_prices' ][ component.component_id ],
				price         = price_data[ 'prices' ][ component.component_id ] + price_data[ 'addons_prices' ][ component.component_id ],
				totals        = {
					price:          0.0,
					regular_price:  0.0,
					price_incl_tax: 0.0,
					price_excl_tax: 0.0
				};

			composite.console_log( 'debug:models', 'Calculating "' + component.get_title() + '" totals...' );

			if ( wc_composite_params.calc_taxes === 'yes' ) {

				if ( product_id > 0 && qty > 0 && ( price > 0 || regular_price > 0 ) ) {

					totals = model.get_taxed_totals( price, regular_price, tax_ratios, qty );
				}

			} else {

				totals.price          = qty * price;
				totals.regular_price  = qty * regular_price;
				totals.price_incl_tax = qty * price;
				totals.price_excl_tax = qty * price;
			}

			// Pass through 'component_totals' filter - @see WC_CP_Filters_Manager class.
			return composite.filters.apply_filters( 'component_totals', [ totals, component, qty ] );
		},

		/**
		 * Updates the 'price_data' model property with the latest component prices.
		 */
		update_component_prices: function( component ) {

			composite.console_log( 'debug:models', 'Fetching "' + component.get_title() + '" price data...' );

			var quantity = component.get_selected_quantity();

			// Copy prices.
			this.price_data[ 'prices' ][ component.component_id ]           = component.component_selection_model.get_price();
			this.price_data[ 'regular_prices' ][ component.component_id ]   = component.component_selection_model.get_regular_price();
			this.price_data[ 'price_tax_ratios' ][ component.component_id ] = component.component_selection_model.get_tax_ratios();


			// Calculate addons price.
			this.price_data[ 'addons_prices' ][ component.component_id ]  = Number( component.component_selection_model.get_addons_price() );

			if ( quantity > 0 ) {
				this.price_data[ 'quantities' ][ component.component_id ] = parseInt( quantity, 10 );
			} else {
				this.price_data[ 'quantities' ][ component.component_id ] = 0;
			}
		},

		add_status_message: function( source, content ) {

			var messages = $.extend( true, [], this.get( 'status_messages' ) );

			messages.push( { message_source: source, message_content: content } );

			composite.console_log( 'debug:models', 'Adding "' + source + '" status message: "' + content + '"...' );

			this.set( { status_messages: messages } );
		},

		remove_status_message: function( source ) {

			composite.console_log( 'debug:models', 'Removing "' + source + '" status message...' );

			var messages = _.filter( this.get( 'status_messages' ), function( status_message ) { return status_message.message_source !== source; } );

			this.set( { status_messages: messages } );
		}

	} );

	var obj = new Model( opts );
	return obj;
};
